//LICENSE Portions Copyright 2019-2021 ZomboDB, LLC.
//LICENSE
//LICENSE Portions Copyright 2021-2023 Technology Concepts & Design, Inc.
//LICENSE
//LICENSE Portions Copyright 2023-2023 PgCentral Foundation, Inc. <contact@pgcentral.org>
//LICENSE
//LICENSE All rights reserved.
//LICENSE
//LICENSE Use of this source code is governed by the MIT license that can be found in the LICENSE file.
/*!

`#[derive(PostgresHash)]` related entities for Rust to SQL translation

> Like all of the [`sql_entity_graph`][crate] APIs, this is considered **internal**
> to the `pgrx` framework and very subject to change between versions. While you may use this, please do it with caution.

*/
use crate::pgrx_sql::PgrxSql;
use crate::to_sql::ToSql;
use crate::to_sql::entity::ToSqlConfigEntity;
use crate::{SqlGraphEntity, SqlGraphIdentifier};

/// The output of a [`PostgresHash`](crate::postgres_hash::PostgresHash) from `quote::ToTokens::to_tokens`.
#[derive(Debug, Clone, Hash, PartialEq, Eq, Ord, PartialOrd)]
pub struct PostgresHashEntity<'a> {
    pub name: &'a str,
    pub file: &'a str,
    pub line: u32,
    pub full_path: &'a str,
    pub module_path: &'a str,
    pub type_ident: &'a str,
    pub to_sql_config: ToSqlConfigEntity<'a>,
}

impl PostgresHashEntity<'_> {
    pub(crate) fn fn_name(&self) -> String {
        format!("{}_hash", self.name.to_lowercase())
    }
}

impl<'a> From<PostgresHashEntity<'a>> for SqlGraphEntity<'a> {
    fn from(val: PostgresHashEntity<'a>) -> Self {
        SqlGraphEntity::Hash(val)
    }
}

impl SqlGraphIdentifier for PostgresHashEntity<'_> {
    fn dot_identifier(&self) -> String {
        format!("hash {}", self.full_path)
    }
    fn rust_identifier(&self) -> String {
        self.full_path.to_string()
    }

    fn file(&self) -> Option<&str> {
        Some(self.file)
    }

    fn line(&self) -> Option<u32> {
        Some(self.line)
    }
}

impl ToSql for PostgresHashEntity<'_> {
    fn to_sql(&self, context: &PgrxSql) -> eyre::Result<String> {
        let self_index = context.hashes[self];
        let schema = context.schema_prefix_for(&self_index);
        let sql = format!(
            "\n\
                            -- {file}:{line}\n\
                            -- {full_path}\n\
                            CREATE OPERATOR FAMILY {schema}{name}_hash_ops USING hash;\n\
                            CREATE OPERATOR CLASS {schema}{name}_hash_ops DEFAULT FOR TYPE {schema}{name} USING hash FAMILY {schema}{name}_hash_ops AS\n\
                                \tOPERATOR    1   =  ({schema}{name}, {schema}{name}),\n\
                                \tFUNCTION    1   {schema}{fn_name}({schema}{name});\
                            ",
            name = self.name,
            full_path = self.full_path,
            file = self.file,
            line = self.line,
            fn_name = self.fn_name(),
        );
        Ok(sql)
    }
}
